#    Copyright 2008 Peter Bulychev
#    http://clonedigger.sourceforge.net
#
#    This file is part of Clone Digger.
#
#    Clone Digger is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Clone Digger is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with Clone Digger.  If not, see <http://www.gnu.org/licenses/>.


class SuffixTree:
    class StringPosition:
        def __init__(self, string, position, prevelem):
            self.string = string
            self.position = position
            self.prevelem = prevelem

    class SuffixTreeNode:
        def __init__(self):
            self.childs = {}
            self.string_positions = []
            self.ending_strings = []

    def __init__(self, f_code):
        self._node = self.SuffixTreeNode()
        self._f_code = f_code

    def _add(self, string, prevelem):
        pos = 0
        node = self._node
        for pos in range(len(string)):
            e = string[pos]
            code = self._f_code(e)
            node.string_positions.append(self.StringPosition(string, pos, prevelem))
            if code not in node.childs:
                node.childs[code] = self.SuffixTreeNode()
            node = node.childs[code]
        node.ending_strings.append(self.StringPosition(string, pos+1, prevelem))

    def add(self, string):
        for i in range(len(string)):
            if i == 0:
                prevelem = None
            else:
                prevelem = self._f_code(string[i-1])
            self._add(string[i:], prevelem)

    def getBestMaxSubstrings(self, threshold, f, f_elem, node=None, initial_threshold=None):
        if initial_threshold is None:
            initial_threshold = threshold

        def check_left_diverse_and_add(s1, s2, p):
            if ((s1.prevelem is None) or (s2.prevelem is None) or (s1.prevelem != s2.prevelem)) and s1.position > p:
                candidate = (s1.string[:s1.position-p], s2.string[:s2.position-p])
                if f_elem(candidate[0]) >= initial_threshold or \
                   f_elem(candidate[1]) >= initial_threshold:
                    r.append(candidate)
                return True
            else:
                return False
        if node is None:
            node = self._node
        r = []
        if threshold <= 0:
            for s1 in node.ending_strings:
                for s2 in node.string_positions:
                    if s1.string == s2.string:
                        continue
                    check_left_diverse_and_add(s1, s2, 0)
            for i in range(len(node.ending_strings)):
                for j in range(i):
                    s1 = node.ending_strings[i]
                    s2 = node.ending_strings[j]
                    check_left_diverse_and_add(s1, s2, 0)
            for i in range(len(node.childs.keys())):
                for j in range(i):
                    c1 = node.childs.keys()[i]
                    c2 = node.childs.keys()[j]
                    for s1 in node.childs[c1].string_positions + node.childs[c1].ending_strings:
                        for s2 in node.childs[c2].string_positions + node.childs[c2].ending_strings:
                            check_left_diverse_and_add(s1, s2, 1)
        for (code, child) in node.childs.items():
            r += self.getBestMaxSubstrings(threshold - f(code), f, f_elem, child, initial_threshold)
        return r

if __name__ == '__main__':
    class Elem:
        def __init__(self, code):
            self._code = code

        def getCode(self):
            return self._code

        def __str__(self):
            return str(self._code)

    def test1():
        t = SuffixTree()
        for w in ['abcPeter', 'Pet1erbca', 'Peter', 'aPet0--']:
            t.add([Elem(c) for c in w])
        maxs = t.getBestMaxSubstrings(3)
        l = []
        for (s1, s2) in maxs:
            l.append([''.join([str(e) for e in s1]), ''.join([str(e) for e in s2])])
        assert(l == [['Pe1t', 'P2et'], ['P3et', 'Pe4t'], ['Pet', 'Pet'], ['Pet', 'Pet'], ['Pet', 'Pet'], ['Peter', 'Peter']])

    def test2():
        t = SuffixTree()
        for w in ['a', 'aa']:
            t.add([Elem(c) for c in w])
        maxs = t.getBestMaxSubstrings(0)
        l = []
        for (s1, s2) in maxs:
            l.append([''.join([str(e) for e in s1]), ''.join([str(e) for e in s2])])
        assert(l == [['a', 'a'], ['a', 'a'], ['a', 'a']])
    for s in dir():
        if s.find('test') == 0:
            eval(s + '()')



#    Copyright 2008 Peter Bulychev
#
#    This file is part of Clone Digger.
#
#    Clone Digger is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Clone Digger is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with Clone Digger.  If not, see <http://www.gnu.org/licenses/>.

import copy

from abstract_syntax_tree import *

# NOTE that everywhere is written Unifier instead of AntiUnifier, for simplicity

# Constants here
verbose = True


class FreeVariable(AbstractSyntaxTree):
    free_variables_count = 1

    def __init__(self, parameters):
        global free_variables_count
        FreeVariable.free_variables_count += 1
        name = 'VAR(%d)' % (FreeVariable.free_variables_count, )
#       self._childs = []
        AbstractSyntaxTree.__init__(self, parameters, name)


class Substitution:
    def __init__(self, initial_value=None):
        if initial_value is None:
            initial_value = {}
        self._map = initial_value

    def substitute(self, tree, without_copying=False):
        if tree in self._map.keys():
            return self._map[tree]
        else:
            if isinstance(tree, FreeVariable):
                return tree
            if without_copying:
                return tree
            else:
                r = AbstractSyntaxTree(tree._parameters, tree.getName())
                for child in tree.getChilds():
                    r.addChild(self.substitute(child, without_copying))
                return r

    def getMap(self):
        return self._map

    def getSize(self):
        ret = 0
        for (u, tree) in self.getMap().items():
            ret += tree.getSize(False) - free_variable_cost
        return ret


class Unifier:
    def __init__(self, t1, t2, ignore_parametrization=False):
        def combineSubs(node, s, t):
            # s and t are 2-tuples
            assert(s[0].getMap().keys() == s[1].getMap().keys())
            assert(t[0].getMap().keys() == t[1].getMap().keys())
            newt = (copy.copy(t[0]), copy.copy(t[1]))
            relabel = {}
            for si in s[0].getMap().keys():
                if not ignore_parametrization:
                    foundone = False
                    for ti in t[0].getMap().keys():
                        if (s[0].getMap()[si] == t[0].getMap()[ti]) and (s[1].getMap()[si] == t[1].getMap()[ti]):
                            relabel[si] = ti
                            foundone = True
                            break
                if ignore_parametrization or not foundone:
                    newt[0].getMap()[si] = s[0].getMap()[si]
                    newt[1].getMap()[si] = s[1].getMap()[si]
            return (Substitution(relabel).substitute(node), newt)

        def unify(node1, node2):
            parameters = node1._parameters
            if node1 == node2:
                return (node1, (Substitution(), Substitution()))
            elif (node1.getName() != node2.getName()) or (node1.getChildCount() != node2.getChildCount()):
                var = FreeVariable(parameters)
                return (var, (Substitution({var: node1}), Substitution({var: node2})))
            else:
                s = (Substitution(), Substitution())
                name = node1.getName()
                retNode = AbstractSyntaxTree(parameters, name)
                count = node1.getChildCount()
                for i in range(count):
                    (ai, si) = unify(node1.getChilds()[i], node2.getChilds()[i])
                    (ai, s) = combineSubs(ai, si, s)
                    retNode.addChild(ai)
                return (retNode, s)
        (self._unifier, self._substitutions) = unify(t1, t2)
        self._unifier.storeSize()
        for i in (0, 1):
            for key in self._substitutions[i].getMap():
                self._substitutions[i].getMap()[key].storeSize()

    def getSubstitutions(self):
        return self._substitutions

    def getUnifier(self):
        return self._unifier

    def getSize(self):
        return sum([s.getSize() for s in self.getSubstitutions()])


class Cluster:
    count = 0

    def __init__(self, tree=None):
        if tree:
            self._n = 1
            self._unifier_tree = tree
            self._trees = [tree]
            self._max_covered_lines = len(tree.getCoveredLineNumbers())
        else:
            self._n = 0
            self._trees = []
            self._max_covered_lines = 0
        Cluster.count += 1
        self._cluster_number = Cluster.count

    def getUnifierTree(self):
        return self._unifier_tree

    def getCount(self):
        return self._n

    def getAddCost(self, tree):
        unifier = Unifier(self.getUnifierTree(), tree)
        return (self.getCount() * unifier.getSubstitutions()[0].getSize() + unifier.getSubstitutions()[1].getSize())

    def unify(self, tree):
        self._n += 1
        self._unifier_tree = Unifier(self.getUnifierTree(), tree).getUnifier()
        self._trees.append(tree)

    def eraseAllTrees(self):
        self._n = 0
        self._trees = []

    def addWithoutUnification(self, tree):
        self._n += 1
        self._trees.append(tree)
        if len(tree.getCoveredLineNumbers()) > self._max_covered_lines:
            self._max_covered_lines = len(tree.getCoveredLineNumbers())

    def getMaxCoveredLines(self):
        return self._max_covered_lines

    def getUnifierSize(self):
        return self.getUnifierTree().getSize()



#    Copyright 2008 Peter Bulychev
#    Copyright 2015 Will Oursler
#
#    This file is part of Clone Digger.
#
#    Clone Digger is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Clone Digger is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with Clone Digger.  If not, see <http://www.gnu.org/licenses/>.
import os
from jinja2 import Environment, FileSystemLoader

path = os.path.dirname(os.path.abspath(__file__))
env = Environment(
  autoescape=False,
  loader=FileSystemLoader(os.path.join(path, 'templates')),
  trim_blocks=False)
template = env.get_template('template.html')


def write(report, file_name):
  percent_source_line_clones = (not report.all_source_lines_count and 100) or 100*report.covered_source_lines_count/float(report.all_source_lines_count)
  # TODO: Make a lot of these properties of report...
  html_text = template.render(
    report=report,
    percent_source_line_clones=percent_source_line_clones)
  f = open(file_name, 'w')
  f.write(html_text)
  f.close()



#!/usr/bin/python

#    Copyright 2008 Peter Bulychev
#    Copyright 2015 Will Oursler
#
#    This file is part of Clone Digger.
#
#    Clone Digger is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Clone Digger is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with Clone Digger.  If not, see <http://www.gnu.org/licenses/>.
import sys

import os
import traceback
import click
import pickle  # TODO: THIS IS A SECURITY RISK. Make JSON based version.
from tabulate import tabulate

import ast_suppliers
import clone_detection_algorithm

from parameters import Parameters
from report import Report, CloneSummary, Snippet
import html_writer

# TODO: Incorprate into CLI?
# TODO: Rewrite CLI as calls to API, once API exists.
"""To run Clone Digger type:
python clonedigger.py [OPTION]... [SOURCE FILE OR DIRECTORY]...

The typical usage is:
python clonedigger.py source_file_1 source_file_2 ...
  or
python clonedigger.py path_to_source_tree
Don't forget to remove automatically generated sources, tests and third party libraries from the source tree.

Notice:
The semantics of threshold options is discussed in the paper "Duplicate code detection using anti-unification", which can be downloaded from the site http://clonedigger.sourceforge.net . All arguments are optional. Supported options are:
"""

# TODO: Remaining stuff... Add or replace functionality with .files
# TODO: Removing prefix banning for now.
'''
  cmdline.add_option('-f', '--force',
             action='store_true', dest='force',
             help='')
  cmdline.add_option('--force-diff',
             action='store_true', dest='use_diff',
             help='force highlighting of differences based on the diff algorithm')
  cmdline.add_option('--fast',
             action='store_true', dest='clusterize_using_hash',
             help='find only clones, which differ in variable and function names and constants')
  cmdline.add_option('--ignore-dir',
             action='append', dest='ignore_dirs',
             help='exclude directories from parsing')
  cmdline.add_option('--report-unifiers',
             action='store_true', dest='report_unifiers',
             help='')
  cmdline.add_option('--func-prefixes',
             action='store',
             dest='f_prefixes',
             help='skip functions/methods with these prefixes (provide a CSV string as argument)')
  cmdline.add_option('--file-list', dest='file_list',
             help='a file that contains a list of file names that must be processed by Clone Digger')

######################

@click.option('--clusterize-using-dcup',
              is_flag=True,
              help="Mark each statement with its D-cup value instead of the most similar pattern.")
'''


@click.group()
def orphanblack_cli():
    pass


@orphanblack_cli.command()
@click.option('-l', '--language',
              type=click.Choice(['python', 'java', 'lua', 'javascript', 'js']),
              default='python',
              help="The language of the provided files.")
@click.option('--no-recursion', is_flag=True)
@click.option('--distance-threshold',
              type=int,
              default=None)  # TODO: Help
@click.option('--hashing-depth',
              type=int,
              default=None)  # TODO: Help
@click.option('--size-threshold',
              type=int,
              default=None)  # TODO: Help
@click.argument('source_file_names',
                type=click.Path(exists=True),
                nargs=-1)
def scan(language, no_recursion, distance_threshold, hashing_depth, size_threshold, source_file_names):

  supplier = ast_suppliers.abstract_syntax_tree_suppliers[language]

  parameters = Parameters()

  if distance_threshold is None:
    parameters.distance_threshold = supplier.distance_threshold
  else:
    parameters.distance_threshold = distance_threshold

  if size_threshold is None:
    parameters.size_threshold = supplier.size_threshold
  else:
    parameters.size_threshold = size_threshold

  source_files = []
  source_file_names = list(source_file_names)

  report = Report(parameters)

  ####### TODO: MAKE FILE LIST
  ####### TODO: Populate parameters
  #for option in cmdline.option_list:
  #  if option.dest == 'file_list' and options.file_list is not None:
  #    source_file_names.extend(open(options.file_list).read().split())
  #    continue
  #  elif option.dest is None:
  #    continue
  #  setattr(arguments, option.dest, getattr(options, option.dest))
  ########

  def parse_file(file_name):
    try:
      print 'Parsing ', file_name, '...',
      sys.stdout.flush()
      source_file = supplier(file_name, parameters)
      source_file.getTree().propagateCoveredLineNumbers()
      source_file.getTree().propagateHeight()
      source_files.append(source_file)
      report.addFileName(file_name)
      print 'done'
    except:
      s = 'Error: can\'t parse "%s" \n: ' % (file_name,) + traceback.format_exc()
      report.addErrorInformation(s)
      print s

  def walk(dirname):
    for dirpath, dirs, files in os.walk(file_name):
      dirs[:] = (not ignore_dirs and dirs) or [d for d in dirs if d not in options.ignore_dirs]
      # Skip all non-parseable files
      files[:] = [f for f in files
                  if os.path.splitext(f)[1][1:] == supplier.extension]
      yield (dirpath, dirs, files)

  for file_name in source_file_names:
    if os.path.isdir(file_name):
      if no_recursion:
        dirpath = file_name
        files = [os.path.join(file_name, f) for f in os.listdir(file_name)
                 if os.path.splitext(f)[1][1:] == supplier.extension]
        for f in files:
          parse_file(f)
      else:
        for dirpath, dirnames, filenames in walk(file_name):
          for f in filenames:
            parse_file(os.path.join(dirpath, f))
    else:
      parse_file(file_name)

  duplicates = clone_detection_algorithm.findDuplicateCode(source_files, report)
  n = 1
  for duplicate in duplicates:
    distance = duplicate.calcDistance()
    summary = CloneSummary(
      "Clone #"+str(n),
      [  # TODO: This is a mess! Most of this info should be assembled on the fly and in member functions.
       Snippet(
        duplicate[i].getSourceFile()._file_name,
        duplicate[i].getCoveredLineNumbers(),
        '\n'.join([line for line in duplicate[i].getSourceLines()])
        ) for i in [0, 1]], distance)
    report.addClone(summary)
    n += 1
  report.sortByCloneSize()

  with open(".orphanblack", "wb") as f:
    pickle.dump(report, f)


@orphanblack_cli.command()
@click.option('-v', '--verbose', is_flag=True)
def report(verbose):
  with open(".orphanblack", "r") as f:
    report = pickle.load(f)
  print "Found", len(report.clones), "clones.\n\n"
  for clone in report.clones:
    print clone.name
    print "="*len(clone.name)
    print "Distance between two fragments =", clone.distance
    print "Clone size =", clone.size

    def rangify(lines):
      return str(min(lines)) + ' through ' + str(max(lines))

    table = [[snippet.filename, str(snippet.first_line) + ' through ' + str(snippet.last_line)] for snippet in clone.snippets]
    print tabulate(table, headers=["File", "Lines"], tablefmt="fancy_grid")
    if verbose:
      for snippet in clone.snippets:
        print ''
        title = snippet.filename + ":" + str(snippet.first_line) + "-" + str(snippet.last_line)
        print "*"+"="*len(title)+"*"
        print "|"+title+"|"
        print "*"+"="*len(title)+"*"
        print snippet.text
    print '\n\n'


@orphanblack_cli.command()
@click.option('-o', '--output',
              'output_file_name',
              type=click.Path(),
              default='output.html',
              help="An HTML report will be written to this file. \
                    Defaults to output.html")
def html(output_file_name):
  """Outputs a readable html page."""
  with open(".orphanblack", "r") as f:
    report = pickle.load(f)
  html_writer.write(report, output_file_name)

# This portion of the CLI implements copyright and liscense notices in line
# with the GNU GPL3 best practices. It is not a replacement for LICENSE.txt.

copyright_message = """
orphanblack is a refactor (derivative work) of clonedigger

orphanblack   Copyright (C) 2015 Will Oursler
clonedigger   Copyright (C) 2008 Peter Bulychev

orphanblack comes with ABSOLUTELY NO WARRANTY; for details type `show warranty'.

This is free software, and you are welcome to redistribute it under certain
conditions; see LICENSE.txt for details.
"""

warranty_message = """
orphanblack is licensed under the GNU GPL 3.

As such, there is no warranty for the program, to the extent permitted by
applicable law.  Except when otherwise stated in writing the copyright
holders and/or other parties provide the program "as is" without warranty
of any kind, either expressed or implied, including, but not limited to,
the implied warranties of merchantability and fitness for a particular
purpose.  The entire risk as to the quality and performance of the program
is with you.  Should the program prove defective, you assume the cost of
all necessary servicing, repair or correction.

In no event unless required by applicable law or agreed to in writing
will any copyright holder, or any other party who modifies and/or conveys
the program as permitted above, be liable to you for damages, including any
general, special, incidental or consequential damages arising out of the
use or inability to use the program (including but not limited to loss of
data or data being rendered inaccurate or losses sustained by you or third
parties or a failure of the program to operate with any other programs),
even if such holder or other party has been advised of the possibility of
such damages.

For more details, see the full license, which should have been distributed
with this software as "LICENSE.txt".
"""


@orphanblack_cli.group()
def show():
  pass


@show.command()
def copyright():
  print copyright_message


@show.command()
def warranty():
  print warranty_message


if __name__ == '__main__':
  orphanblack_cli()



# TODO: report serialization without pickle because that's insecure.

import textwrap


# TODO: TEXT OF FILES!
class Snippet:
  def __init__(self, filename, lines, text):
    def deindent(string):
      if string and string[0] == '\n':
        string = string[1:]
      return textwrap.dedent(string)
    self.filename = filename
    self.lines = lines
    self.text = deindent(text)

  @property
  def first_line(self):
    return min(self.lines)

  @property
  def last_line(self):
    return max(self.lines)

  @property
  def size(self):
    return len(self.lines)


# Minimal Clone Summary for serialization.
class CloneSummary:
  def __init__(self, name, snippets, distance):
    self.name = name
    self.snippets = snippets
    self.distance = distance

  @property
  def size(self):
    return sum([snippet.size for snippet in self.snippets]) / len(self.snippets)


class Report:
  def __init__(self, parameters):
    self._parameters = parameters
    self._error_info = []
    self._clones = []
    self._file_names = []
    self._mark_to_statement_hash = None

  @property
  def parameters(self):
    return self._parameters

  @property
  def clones(self):
    return self._clones

  @property
  def filenames(self):
    return self._file_names

  def setMarkToStatementHash(self, mark_to_statement_hash):  # TODO: Figure out what this is and if it belongs
    self._mark_to_statement_hash = mark_to_statement_hash

  def addFileName(self, file_name):
    self._file_names.append(file_name)

  def addErrorInformation(self, error_info):
    self._error_info.append(error_info)

  def addClone(self, clone):
    self._clones.append(clone)

  def sortByCloneSize(self):
    def f(a, b):
      return cmp(b.size, a.size)
    self._clones.sort(f)



#    Copyright 2008 Peter Bulychev
#    Copyright 2015 Will Oursler
#
#    This file is part of Clone Digger.
#
#    Clone Digger is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Clone Digger is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with Clone Digger.  If not, see <http://www.gnu.org/licenses/>.

import types

free_variable_cost = 0.5


class ParseError:
    def __init__(self, descr):
        self._descr = descr

    def __str__(self):
        return self._descr


class SourceFile:
    size_threshold = 5
    distance_threshold = 5

    def __init__(self, file_name, parameters):
        f = open(file_name, 'r')

        def filter_func(s):
            for i in range(len(s)-1, -2, -1):
                if i < 0 or not s[i].isspace():
                    break
            if i >= 0:
                return s[:i+1]
            else:
                return s
        self._source_lines = [filter_func(s) for s in f.readlines()]
        f.close()
        self._file_name = file_name
        self._parameters = parameters

    def getSourceLine(self, n):
#       if n >= len(self._source_lines):
#           return ''
#TODO
#error here
        return self._source_lines[n]

    def _setTree(self, tree):
        self._tree = tree

    def getTree(self):
        return self._tree

    def getFileName(self):
        return self._file_name


class AbstractSyntaxTree:
    def __init__(self, parameters, name, line_numbers=[], source_file=None):
        self._parameters = parameters
        self._childs = []
        self._line_numbers = line_numbers
        self._covered_line_numbers = None
        self._parent = None
        self._hash = None
        self._source_file = source_file
        self._is_statement = False
        self._name = name

    def getSourceFile(self):
        return self._source_file

    def setMark(self, mark):
        self._mark = mark

    def getMark(self):
        return self._mark

    def markAsStatement(self, val=True):
        self._is_statement = val

    def isStatement(self):
        return self._is_statement

    def setName(self, name):
        self._name = name

    def getLineNumbers(self):
        return self._line_numbers

    def getCoveredLineNumbers(self):
        return self._covered_line_numbers

    def getStartLineNumber(self):
        return min(self.getCoveredLineNumbers())

    def getParent(self):
        return self._parent

    def setParent(self, parent):
        self._parent = parent

    def getAncestors(self):
        r = []
        t = self.getParent()
        while t:
            if t.isStatement():
                r.append(t)
            t = t.getParent()
        return r

    def getSourceLines(self):
        source_line_numbers = set([])
        r = []
        source_line_numbers = self.getCoveredLineNumbers()
        source_line_numbers_list = list(range(min(source_line_numbers), max(source_line_numbers)+1))
        source_line_numbers_list.sort()
        for source_line_number in source_line_numbers_list:
            r.append(self.getSourceFile().getSourceLine(source_line_number))
        return r

    def getName(self):
        return self._name

    def getChilds(self):
        return self._childs

    def getChildCount(self):
        return len(self._childs)

    def propagateCoveredLineNumbers(self):
        self._covered_line_numbers = set(self._line_numbers)
        for child in self.getChilds():
            self._covered_line_numbers.update(child.propagateCoveredLineNumbers())
        return self._covered_line_numbers

    def propagateHeight(self):
        if self.getChildCount() == 0:
            self._height = 0
        else:
            self._height = max([c.propagateHeight() for c in self.getChilds()])+1
        return self._height

    def getHeight(self):
        return self._height

    def addChild(self, child, save_parent=False):
        if not save_parent:
            child.setParent(self)
        self._childs.append(child)

    def setChildCount(self, count):
        assert(not self._childs)
        self._childs = count * [None]

    def setNextUndefinedChild(self, c):
        for i in range(len(self.getChilds())):
            if self.getChilds()[i] is None:
                self._childs[i] = c
        assert()

    def __str__(self):
        return ' ( ' + self.getName() + ' '.join([str(child) for child in self.getChilds()]) + ' ) '

    def getFullHash(self):
        return self.getDCupHash(-1)

    def getDCupHash(self, level):
        if len(self._childs) == 0:
            ret = 0  # in case of names and constants
        else:
            ret = (level+1) * hash(self._name) * len(self._childs)
        # if level == -1, it will not stop until it reaches the leaves
        if level != 0:
            for i in range(len(self._childs)):
                child = self._childs[i]
                ret += (i+1)*child.getDCupHash(level-1)
        return hash(ret)

    def __hash__(self):
        #TODO check correctness
        if not self._hash:
            self._hash = hash(self.getDCupHash(3) + hash(self.getName()))
        return self._hash
#       return  hash(self.getDCupHash(3) + hash(self.getName()))

    def __eq__(self, tree2):
        tree1 = self
        if type(tree2) == types.NoneType:
            return False
        if tree1.getName() != tree2.getName():
            return False
        if tree1.getChildCount() != tree2.getChildCount():
            return False
        for i in range(tree1.getChildCount()):
            if tree1.getChilds()[i] != tree2.getChilds()[i]:
                return False
        return True

    def getAllStatementSequences(self):
        r = []
        current = StatementSequence(self._parameters)
        for child in self.getChilds():
            if child.isStatement():
                current.addStatement(child)
            else:
                if (not current.isEmpty()) and len(current.getCoveredLineNumbers()) >= self._parameters.size_threshold:
                    r.append(current)
                    current = StatementSequence(self._parameters)
            r.extend(child.getAllStatementSequences())
        if (not current.isEmpty()) and len(current.getCoveredLineNumbers()) >= self._parameters.size_threshold:
            r.append(current)
        return r

    def storeSize(self):
        observed = set()
        self._none_count = 0

        def rec_calc_size(t):
            r = 0
            if not t in observed:
                if t.getChildCount():
                    for c in t.getChilds():
                        r += rec_calc_size(c)
                else:
                    observed.add(t)
                    if t.getName() == 'None':
                        self._none_count += 1
                    if t.__class__.__name__ == 'FreeVariable':
                        r += free_variable_cost
                    else:
                        r += 1
            return r
        if not hasattr(self, '_size'):
            self._size = rec_calc_size(self)

    def getSize(self, ignore_none=True):
        ret = self._size
        if ignore_none:
            ret -= self._none_count
        return ret

    def getTokenCount(self):
        def rec_calc_size(t):
            if t.getChildCount():
                if t.getName() in ['Add', 'Assign', 'Sub', 'Div', 'Mul', 'Mod', 'Function', 'If', 'Class', 'Raise']:
                    r = 1
                else:
                    r = 0
                for c in t.getChilds():
                    r += rec_calc_size(c)
            else:
                if t.getName()[0] != "'" and t.getName() != 'Pass':
                   return 0
                else:
                   return 1
            return r
        return rec_calc_size(self)


class StatementSequence:
    def __init__(self, parameters, sequence=[]):
        self._parameters = parameters
        self._sequence = []
        self._source_file = None
        for s in sequence:
            self.addStatement(s)

    def getCoveredLineNumbers(self):
        r = set()
        for s in self:
            r.update(s.getCoveredLineNumbers())
        return r

    def getStartLineNumber(self):
        return min(self.getCoveredLineNumbers())

    def getAncestors(self):
        return self[0].getAncestors()

    def isEmpty(self):
        return (self._sequence == [])

    def addStatement(self, statement):
        self._sequence.append(statement)
        if self._source_file is None:
            self._source_file = statement.getSourceFile()
        else:
            assert(self._source_file == statement.getSourceFile())

    def __getitem__(self, *args):
        return self._sequence.__getitem__(*args)

    def __len__(self):
        return self._sequence.__len__()

    def __str__(self):
        return ','.join([str(s) for s in self])

    def getWeight(self):
        return sum([s.getCluster().getUnifierSize() for s in self._sequence])

    def getSourceFile(self):
        return self._source_file

    def getSourceLines(self):
        source_line_numbers = set([])
        r = []
        for statement in self:
            r.extend(statement.getSourceLines())
        return r

    def getLineNumbers(self):
        r = []
        for statement in self:
            r.extend(statement.getLineNumbers())
        return r

    def getLineNumberHashables(self):
        source_file_name = self.getSourceFile().getFileName()
        line_numbers = self.getCoveredLineNumbers()
        return set([(source_file_name, line_number) for line_number in line_numbers])

    def constructTree(self):
        tree = AbstractSyntaxTree(self._parameters, name='__SEQUENCE__')
        for statement in self:
            tree.addChild(statement, True)
        return tree

    def getLength(self):
        return len(self)

    def getCoveredLineNumbersCount(self):
        covered = set()
        for t in self:
            covered.update(t.getCoveredLineNumbers())
        return len(covered)


class PairSequences:
    def __init__(self, parameters, sequences):
        self._parameters = parameters
        self._sequences = sequences

    def __getitem__(self, *args):
        return self._sequences.__getitem__(*args)

    def __str__(self):
        return ';\t'.join([str(s) for s in self])

    def getWeight(self):
        assert(self[0].getWeight() == self[1].getWeight())
        return self[0].getWeight()

    def calcDistance(self):
        import anti_unification
        trees = [s.constructTree() for s in self]
        unifier = anti_unification.Unifier(trees[0], trees[1])
        return unifier.getSize()

    def calcSize(self):
        return max([len(set(self[i].getCoveredLineNumbers())) for i in [0, 1]])

    def subSequence(self, first, length):
        return PairSequences(self._parameters, [StatementSequence(self._parameters, self[0][first:first+length]), StatementSequence(self._parameters, self[1][first:first+length])])

    def getLength(self):
        return self[0].getLength()

    def getMaxCoveredLineNumbersCount(self):
        return min([s.getCoveredLineNumbersCount() for s in self])



#    Copyright 2008 Peter Bulychev
#    http://clonedigger.sourceforge.net
#
#    This file is part of Clone Digger.
#
#    Clone Digger is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Clone Digger is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with Clone Digger.  If not, see <http://www.gnu.org/licenses/>.

import os
import xml.parsers.expat

from abstract_syntax_tree import *


class JsANTLRSourceFile (SourceFile):
    extension = 'js'
    size_threshold = 5
    distance_threshold = 5

    def __init__(self, file_name):
        SourceFile.__init__(self, file_name)

        class ExpatHandler:
            def __init__(self, start_node, parent):
                self.parent = parent
                self.stack = [start_node]

            def start_element(expat_self, xml_node_name, attrs):
                line_number = int(attrs["line_number"])-1
                line_numbers = [line_number]
                if line_numbers == [-1]:
                    line_numbers = []
                name = attrs["name"]
                r = AbstractSyntaxTree(name, line_numbers, self)
                if xml_node_name == "statement_node":
                #if name in ["CALL", "BLOCK"]:
                    r.markAsStatement()
                else:
                    assert(xml_node_name == "node")
                expat_self.stack[-1].addChild(r)
                expat_self.stack.append(r)

            def end_element(self, name):
                self.stack.pop()

        tree_file_name = 'temporary_ast.xml'
        producer_class_path = os.path.join('.', 'js_antlr', 'TreeProducer.jar')
        antlr_class_path = os.path.join('.', 'antlr_runtime', 'antlr-3.1.1.jar')
        if os.name in ['mac', 'posix']:
            class_path_delimeter = ':'
        elif os.name in ['nt', 'dos', 'ce']:
            class_path_delimeter = ';'
        else:
            print 'unsupported OS'
            assert(0)

        if os.system('java -classpath ' + producer_class_path + class_path_delimeter + antlr_class_path + ' TreeProducer %s %s 2>err.log' % (file_name, tree_file_name)):
            f = open('err.log')
            s = f.read()
            f.close()
            raise Exception(s)
        f = open('err.log')
        s = f.read()
        f.close()
        if s:
            print s

        self._tree = AbstractSyntaxTree('program')
        handler = ExpatHandler(self._tree, self)
        p = xml.parsers.expat.ParserCreate()
        p.StartElementHandler = handler.start_element
        p.EndElementHandler = handler.end_element
        f = open(tree_file_name)
        p.ParseFile(f)
        f.close()
#       os.remove(tree_file_name)



# TODO: This file needs to allow explicitly the following:
# Run a scan on a particular path with any file selection criteria
# Load results of any previous scan
# Automatically find canidate results to load.



class Parameters:
  def __init__(self):
    self.clustering_threshold = 10
    self.size_threshold = 5  # ONLY TRUE FOR PYTHON
    self.distance_threshold = 5  # ONLY TRUE FOR PYTHON
    self.hashing_depth = 1
    self.clusterize_using_dcup = False
    self.report_unifiers = False
    self.force = False
    self.use_diff = False
    self.clusterize_using_hash = False



from distutils.core import setup
import os

# This attempts to provide a safe reStructuredTxt for PyPI based on README.md
long_description = 'Static Analysis for Software Clone Detection'
if os.path.exists('README.rst'):
    long_description = open('README.rst').read()

setup(
  name='orphanblack',
  packages=['orphanblack'],
  install_requires=[
    "Jinja2",
    "tabulate",
    "click"
    ],  # TODO: versioning info?
  version='0.1.11',
  description='Static Analysis for Software Clone Detection',
  long_description=long_description,
  author='Will Oursler',
  author_email='woursler@gmail.com',
  url='https://github.com/Zomega/orphanblack',
  download_url='https://github.com/Zomega/orphanblack/tarball/0.1.11',
  keywords=['testing', 'clone', 'static analysis', 'clonedigger', 'buildbot'],
  classifiers=[],
  package_dir={'orphanblack': 'orphanblack'},
  package_data={'orphanblack': ['templates/*.html']},
  entry_points={
    'console_scripts': [
      'orphanblack = orphanblack.cli:orphanblack_cli',
    ]
  }
)



#    Copyright 2008 Peter Bulychev
#
#    This file is part of Clone Digger.
#
#    Clone Digger is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Clone Digger is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with Clone Digger.  If not, see <http://www.gnu.org/licenses/>.

# TODO: JESUS. How many files like this are there?

import sys

import suffix_tree
from anti_unification import *
from abstract_syntax_tree import *


def findDuplicateCode(source_files, report):
    statement_sequences = []
    statement_count = 0
    sequences_lengths = []
    for source_file in source_files:
        sequences = source_file.getTree().getAllStatementSequences()
        statement_sequences.extend(sequences)
        sequences_lengths.extend([len(s) for s in sequences])
        statement_count += sum([len(s) for s in sequences])

    if not sequences_lengths:
        print 'Input is empty or the size of the input is below the size threshold'
        sys.exit(0)

    if verbose:
        n_sequences = len(sequences_lengths)
        avg_seq_length = sum(sequences_lengths)/float(n_sequences)
        max_seq_length = max(sequences_lengths)

        print '%d sequences' % (n_sequences,)
        print 'average sequence length: %f' % (avg_seq_length,)
        print 'maximum sequence length: %d' % (max_seq_length,)
        sequences_without_restriction = statement_sequences
        sequences = []
        if not report.parameters.force:
            for sequence in sequences_without_restriction:
                if len(sequence) > 1000:
                    first_statement = sequence[0]
                    print
                    print '-----------------------------------------'
                    print 'Warning: sequences of statements, consists of %d elements is too long.' % (len(sequence),)
                    print 'It starts at %s:%d.' % (first_statement.getSourceFile().getFileName(), min(first_statement.getCoveredLineNumbers()))
                    print 'It will be ignored. Use --force to override this restriction.'
                    print 'Please refer to http://clonedigger.sourceforge.net/documentation.html'
                    print '-----------------------------------------'
                else:
                    sequences.append(sequence)

    def calc_statement_sizes():
        for sequence in statement_sequences:
            for statement in sequence:
                statement.storeSize()

    def build_hash_to_statement(dcup_hash=True):
        hash_to_statement = {}
        for statement_sequence in statement_sequences:
            for statement in statement_sequence:
                if dcup_hash:
                    # 3 - CONSTANT HERE!
                    h = statement.getDCupHash(report.parameters.hashing_depth)
                else:
                    h = statement.getFullHash()
                if not h in hash_to_statement:
                    hash_to_statement[h] = [statement]
                else:
                    hash_to_statement[h].append(statement)
        return hash_to_statement

    def build_unifiers(hash_to_statement):
        processed_statements_count = 0
        clusters = []
        ret = {}
        for h in hash_to_statement.keys():
            local_clusters = []
            statements = hash_to_statement[h]
            for statement in statements:
                processed_statements_count += 1
                if verbose and ((processed_statements_count % 1000) == 0):
                    print '%d,' % (processed_statements_count,),
                    sys.stdout.flush()
                bestcluster = None
                mincost = sys.maxint
                for cluster in local_clusters:
                    cost = cluster.getAddCost(statement)
                    if cost < mincost:
                        mincost = cost
                        bestcluster = cluster
                assert(local_clusters == [] or bestcluster)
                if mincost < 0:
                    pdb.set_trace()
                assert(mincost >= 0)
                if bestcluster is None or mincost > report.parameters.clustering_threshold:
                    newcluster = Cluster(statement)
                    local_clusters.append(newcluster)
                else:
                    bestcluster.unify(statement)
            ret[h] = local_clusters
            clusters.extend(local_clusters)
        return ret

    def clusterize(hash_to_statement, clusters_map):
        processed_statements_count = 0
        # clusters_map contain hash values for statements, not unifiers
        # therefore it will work correct even if unifiers are smaller than hashing depth value
        for h in hash_to_statement.keys():
            clusters = clusters_map[h]
            for statement in hash_to_statement[h]:
                processed_statements_count += 1
                if verbose and ((processed_statements_count % 1000) == 0):
                    print '%d,' % (processed_statements_count,),
                    sys.stdout.flush()
                mincost = sys.maxint
                for cluster in clusters:
                    new_u = Unifier(cluster.getUnifierTree(), statement)
#                   assert(new_u.getSubstitutions()[0].getSize() == 0)
                    cost = new_u.getSize()
                    if cost < mincost:
                        mincost = cost
                        statement.setMark(cluster)
                        cluster.addWithoutUnification(statement)

    def filterOutLongEquallyLabeledSequences(statement_sequences):
        #TODO - refactor, combine with the previous warning
        sequences_without_restriction = statement_sequences
        statement_sequences = []
        for sequence in sequences_without_restriction:
            new_sequence = copy.copy(sequence._sequence)
            current_mark = None
            length = 0
            first_statement_index = None
            flag = False
            for i in range(len(sequence)):
                statement = sequence[i]
                if statement.getMark() != current_mark:
                    if flag is True:
                        flag = False
                    current_mark = statement.getMark()
                    length = 0
                    first_statement_index = i
                else:
                    length += 1
                    if length > 10:
                        new_sequence[i] = None
                        if not flag:
                            for i in range(first_statement_index, i):
                                new_sequence[i] = None
                            first_statement = sequence[first_statement_index]
                            print
                            print '-----------------------------------------'
                            print 'Warning: sequence of statements starting at %s:%d' % (first_statement.getSourceFile().getFileName(), min(first_statement.getCoveredLineNumbers()))
                            print 'consists of many similar statements.'
                            print 'It will be ignored. Use --force to override this restriction.'
                            print 'Please refer to http://clonedigger.sourceforge.net/documentation.html'
                            print '-----------------------------------------'
                            flag = True
            new_sequence = new_sequence + [None]
            cur_sequence = StatementSequence(report.parameters)
            for statement in new_sequence:
                if statement is None:
                    if cur_sequence:
                        statement_sequences.append(cur_sequence)
                        cur_sequence = StatementSequence(report.parameters)
                else:
                    cur_sequence.addStatement(statement)
        return statement_sequences

    def mark_using_hash(hash_to_statement):
        for h in hash_to_statement:
            cluster = Cluster()
            for statement in hash_to_statement[h]:
                cluster.addWithoutUnification(statement)
                statement.setMark(cluster)

    def findHugeSequences():
        def f_size(x):
            return x.getMaxCoveredLines()

        def f_elem(x):
            return StatementSequence(report.parameters, x).getCoveredLineNumbersCount()

        def fcode(x):
            return x.getMark()
        f = f_size
        suffix_tree_instance = suffix_tree.SuffixTree(fcode)
        for sequence in statement_sequences:
            suffix_tree_instance.add(sequence)
        return [PairSequences(report.parameters, [StatementSequence(report.parameters, s1), StatementSequence(report.parameters, s2)]) for (s1, s2) in suffix_tree_instance.getBestMaxSubstrings(report.parameters.size_threshold, f, f_elem)]

    def refineDuplicates(pairs_sequences):
        r = []
        flag = False
        while pairs_sequences:
            pair_sequences = pairs_sequences.pop()

            def all_pairsubsequences_size_n_threshold(n):
                lr = []
                for first in range(0, pair_sequences.getLength()-n+1):
                    new_pair_sequences = pair_sequences.subSequence(first, n)
                    size = new_pair_sequences.getMaxCoveredLineNumbersCount()
                    if size >= report.parameters.size_threshold:
                        lr.append((new_pair_sequences, first))
                return lr
            n = pair_sequences.getLength() + 1
            while 1:
                n -= 1
                if n == 0:
                    break
                new_pairs_sequences = all_pairsubsequences_size_n_threshold(n)
                for (candidate_sequence, first) in new_pairs_sequences:
                    distance = candidate_sequence.calcDistance()
                    if (distance < report.parameters.distance_threshold):
                        r.append(candidate_sequence)
                        if first > 0:
                            pairs_sequences.append(pair_sequences.subSequence(0, first-1))
                        if first+n < pair_sequences.getLength():
                            pairs_sequences.append(pair_sequences.subSequence(first+n, pair_sequences.getLength() - first - n))
                        n += 1
                        flag = True
                        break
                if flag:
                    flag = False
                    break
        return r

    def remove_dominated_clones(clones):
        ret_clones = []
#       def f_cmp(a, b):
#           return a.getLevel().__cmp__(b.getLevel())
#       clones.sort(f_cmp)
        statement_to_clone = {}
        for clone in clones:
            for sequence in clone:
                for statement in sequence:
                    if statement not in statement_to_clone:
                        statement_to_clone[statement] = []
                    statement_to_clone[statement].append(clone)
        for clone in clones:
            ancestors_2 = clone[1].getAncestors()
            flag = True
            for s1 in clone[0].getAncestors():
                if s1 in statement_to_clone:
                    for clone2 in statement_to_clone[s1]:
                        if s1 in clone2[0]:
                            seq = clone2[1]
                        else:
                            assert(s1 in clone2[1])
                            seq = clone2[0]
                        for s2 in seq:
                            if s2 in ancestors_2:
                                flag = False
                                break
                        if not flag:
                            break
                if not flag:
                    break
            if flag:
                ret_clones.append(clone)
        return ret_clones

    if verbose:
        print 'Number of statements: ', statement_count
        print 'Calculating size for each statement...',
        sys.stdout.flush()
    calc_statement_sizes()
    if verbose:
        print 'done'

    if verbose:
        print 'Building statement hash...',
        sys.stdout.flush()
    if report.parameters.clusterize_using_hash:
        hash_to_statement = build_hash_to_statement(dcup_hash=False)
    else:
        hash_to_statement = build_hash_to_statement(dcup_hash=True)
    if verbose:
        print 'done'
        print 'Number of different hash values: ', len(hash_to_statement)

    if report.parameters.clusterize_using_dcup or report.parameters.clusterize_using_hash:
        print 'Marking each statement with its hash value'
        mark_using_hash(hash_to_statement)
    else:
        if verbose:
            print 'Building patterns...',
            sys.stdout.flush()
        clusters_map = build_unifiers(hash_to_statement)
        if verbose:
            print Cluster.count, 'patterns were discovered'
            print 'Choosing pattern for each statement...',
            sys.stdout.flush()
        clusterize(hash_to_statement, clusters_map)
        if verbose:
            print 'done'

    if report.parameters.report_unifiers:
        if verbose:
            print 'Building reverse hash for reporting ...',
            sys.stdout.flush()
        reverse_hash = {}
        for sequence in statement_sequences:
            for statement in sequence:
                mark = statement.getMark()
                if mark not in reverse_hash:
                    reverse_hash[mark] = []
                reverse_hash[mark].append(statement)
        report.setMarkToStatementHash(reverse_hash)
        if verbose:
            print 'done'

    if verbose:
        print 'Finding similar sequences of statements...',
        sys.stdout.flush()

    if not report.parameters.force:
        statement_sequences = filterOutLongEquallyLabeledSequences(statement_sequences)

    duplicate_candidates = findHugeSequences()
    if verbose:
        print len(duplicate_candidates), ' sequences were found'
        print 'Refining candidates...',
        sys.stdout.flush()
    if report.parameters.distance_threshold != -1:
        clones = refineDuplicates(duplicate_candidates)
    else:
        clones = duplicate_candidates
    if verbose:
        print len(clones), 'clones were found'
    if report.parameters.distance_threshold != -1:
        if verbose:
            print 'Removing dominated clones...',
            sys.stdout.flush()
        old_clone_count = len(clones)
        clones = remove_dominated_clones(clones)
        if verbose:
            print len(clones) - old_clone_count, 'clones were removed'

    covered_source_lines = set()
    for clone in clones:
        for sequence in clone:
            covered_source_lines = covered_source_lines.union(sequence.getLineNumberHashables())
    source_lines = set()
    for sequence in statement_sequences:
        source_lines = source_lines.union(sequence.getLineNumberHashables())
    report.all_source_lines_count = len(source_lines)
    report.covered_source_lines_count = len(covered_source_lines)

    return clones



#    Copyright 2008 Peter Bulychev
#    http://clonedigger.sourceforge.net
#
#    This file is part of Clone Digger.
#
#    Clone Digger is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Clone Digger is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with Clone Digger.  If not, see <http://www.gnu.org/licenses/>.

import compiler

from abstract_syntax_tree import *


class PythonNodeLeaf:
    def __init__(self, val):
        self._val = val

    def getVal(self):
        return self._val

    def as_string(self):
        return str(self.getVal())

    def __str__(self):
        return self.as_string()


class PythonCompilerSourceFile (SourceFile):
    extension = 'py'
    distance_threshold = 5
    size_threshold = 5
    ignored_statements = ['Import', 'From']

    def __init__(self, file_name, parameters, func_prefixes=()):
        SourceFile.__init__(self, file_name, parameters)
        self._func_prefixes = func_prefixes

        def rec_build_tree(compiler_ast_node, is_statement=False):
            def flatten(list):
                l = []
                for elt in list:
                    t = type(elt)
                    if t is tuple or t is list:
                        for elt2 in flatten(elt):
                            l.append(elt2)
                    else:
                        l.append(elt)
                return l

            def add_childs(childs):
                assert isinstance(childs, list)
                for child in childs:
                    assert(isinstance(child, compiler.ast.Node))
                    t = rec_build_tree(child, is_statement)
                    if t.getName() in self.ignored_statements:
                        # TODO move it up
                        continue
                    t.setParent(r)
                    r.addChild(t)

            def add_leaf_child(child, name):
                assert not isinstance(child, list)
                assert not isinstance(child, compiler.ast.Node)
                t = AbstractSyntaxTree(self._parameters, repr(child))
                t.setParent(r)
                l = PythonNodeLeaf(child)
                t.ast_node = l
                r.addChild(t)
                setattr(r.ast_node, name, l)
                return t

            def add_leaf_childs(childs, name):
                assert isinstance(childs, list) or isinstance(childs, tuple)
                a = getattr(r.ast_node, name)
                for i in range(len(childs)):
                    child = childs[i]
                    assert(not isinstance(child, compiler.ast.Node))
                    t = AbstractSyntaxTree(self._parameters, repr(child))
                    t.setParent(r)
                    l = PythonNodeLeaf(child)
                    t.ast_node = l
                    r.addChild(t)
                    a[i] = l

            def add_leaf_string_childs(childs):
                assert isinstance(childs, list)
                for child in childs:
                    assert not isinstance(child, compiler.ast.Node)
                    t = AbstractSyntaxTree(self._parameters, repr(child))
                    t.setParent(t)
                    r.addChild(t)

            if isinstance(compiler_ast_node, compiler.ast.Node):
                name = compiler_ast_node.__class__.__name__
                if name == 'Function':
                   for prefix in self._func_prefixes:
                       if compiler_ast_node.name.startswith(prefix):
                           # skip function that matches pattern
                           return AbstractSyntaxTree(self._parameters, 'none')
                if name in ['Function', 'Class']:
                    # ignoring class and function docs
                    compiler_ast_node.doc = None
                if compiler_ast_node.lineno:
                    lines = [compiler_ast_node.lineno-1]
                else:
                    lines = []
                r = AbstractSyntaxTree(self._parameters, name, lines, self)
                r.ast_node = compiler_ast_node
                if is_statement and compiler_ast_node.lineno:
                    r.markAsStatement()
                is_statement = (name == 'Stmt')
                if name == "AssAttr":
                    add_childs([compiler_ast_node.expr])
                    add_leaf_child(compiler_ast_node.attrname, 'attrname')
                    add_leaf_string_childs([compiler_ast_node.flags])
                elif name == "AssName":
                    add_leaf_child(compiler_ast_node.name, 'name')
#                   add_leaf_child(compiler_ast_node.flags, 'flags')
                elif name == "AugAssign":
                    add_childs([compiler_ast_node.node])
                    add_leaf_child(compiler_ast_node.op, 'op')
                    add_childs([compiler_ast_node.expr])
                elif name == "Class":
                    add_leaf_child(compiler_ast_node.name, 'name')
#                   print '>>>>>>>>>>>>>>>>>>>>', flatten(compiler_ast_node.bases)
                    add_childs(flatten(compiler_ast_node.bases))
#                   add_leaf_child(compiler_ast_node.doc, 'doc') we don't want class docs in our tree, do we?
                    add_childs([compiler_ast_node.code])
                elif name == "Compare":
                    add_childs([compiler_ast_node.expr])
                    for i in range(len(compiler_ast_node.ops)):
                        (op, expr) = compiler_ast_node.ops[i]
                        t = add_leaf_child(op, 'op')
                        add_childs([expr])
                        compiler_ast_node.ops[i] = (t.ast_node, expr)
                elif name == "Const":
                    add_leaf_child(repr(compiler_ast_node.value), "value")
#               elif name == "From":
#                   add_leaf_child(compiler_ast_node.modname, "modname")
#                   add_childs(compiler_ast_node.names)
                elif name == "Function":
#                   add_childs(compiler_ast_node.decorators)  FIXME do we need that?
                    add_leaf_child(compiler_ast_node.name, "name")
                    add_leaf_childs(compiler_ast_node.argnames, "argnames")
                    if compiler_ast_node.defaults == ():
                        compiler_ast_node.defaults = []
                    add_childs(compiler_ast_node.defaults)  # TODO incomment and fix
                    add_leaf_string_childs([compiler_ast_node.flags])
#                   add_leaf_child(compiler_ast_node.doc, "doc") same as class docs... we don't need them
                    add_childs([compiler_ast_node.code])
                elif name == "Getattr":
                    add_childs([compiler_ast_node.expr])
                    add_leaf_child(compiler_ast_node.attrname, "attrname")
                elif name == "Global":
                    add_leaf_childs(compiler_ast_node.names, "names")
#               elif name == "Import":
#                   add_leaf_childs(compiler_ast_node.names, "names")
                elif name == "Keyword":
                    add_leaf_child(compiler_ast_node.name, "name")
                    add_childs([compiler_ast_node.expr])
                elif name == "Lambda":
#                   TODO: uncomment and fix
                    add_leaf_childs(compiler_ast_node.argnames, "argnames")
                    if compiler_ast_node.defaults == ():
                        compiler_ast_node.defaults = []
                    add_childs(compiler_ast_node.defaults)
                    add_childs([compiler_ast_node.code])
                elif name == "Name":
                    # the most important one :)
                    add_leaf_child(compiler_ast_node.name, "name")
                else:
                    for c in compiler_ast_node.getChildren():
                        t = rec_build_tree(c, is_statement)
                        if t.getName() in self.ignored_statements:
                            continue
                        t.setParent(r)
                        r.addChild(t)
                return r
            else:
                return AbstractSyntaxTree(self._parameters, repr(compiler_ast_node))
        self._setTree(rec_build_tree(compiler.parseFile(file_name)))
